#include <algorithm>
#include <cstdlib>
#include <fstream>
#include <iostream>
#include <vector>

#include <misc/utils.h>

#include "graph.h"


using namespace CSA;


/*
  Builds an automaton from aligned DNA sequences. Accepted characters are the
  bases ('A', 'C', 'G', 'T'), 'N' denoting the most common character at the
  current position, and ' ' denoting a gap. There must be at least one
  explicitly given base at each position.

  Input format: n lines of m characters each, separated by '\n'.

  If there are gaps in the alignment, the gap cannot be preceded by a
  character that appears at the last position of the gap in some other
  sequence.

  The program considers a context of k bases before and after each position.
  Two bases are merged to a common state, if they

    1) are identical,
    2) occur in the same position in the alignment, and
    3) share a pre-context or a post-context.

  The reason for using post-context is that the automaton must be
  deterministic in reverse direction. Two different states sharing an
  accepting path can lead to an explosion of states in the expanded automaton.
*/

struct Node
{
  usint id;     // Actual node id.
  char label;   // Character value for the node.

  const std::vector<std::string>& alignment;
  usint pos;    // Position in the alingment. Used for sampling.
  usint seq;    // Sequence number.

  std::string context;
  usint len;    // Context length.
  usint end;    // Position of the last character of the context in alignment.

  Node* link;   // Used for determining which nodes will be created.

  Node(const std::vector<std::string>& _alignment, usint _pos, usint _seq, usint _len) :
    id(0), label(_alignment[_pos][_seq]),
    alignment(_alignment), pos(_pos), seq(_seq),
    context(""), len(_len), end(_pos),
    link(this)
  {
    this->buildContext();
//     std::cout << "  Node(" << this->label << ", " << this->context << ")" << std::endl;
  }

  Node(Node* previous, usint _pos) :
    id(0), label(previous->alignment[_pos][previous->seq]),
    alignment(previous->alignment), pos(_pos), seq(previous->seq),
    context(""), len(previous->len), end(previous->end),
    link(this)
  {
    if(previous->context.length() > 1)
    {
      this->context = previous->context.substr(1);
    }

    this->buildContext();
//     std::cout << "  Node(" << this->label << ", " << this->context << ")" << std::endl;
  }

  void buildContext()
  {
    while(this->context.length() < this->len)
    {
      if(this->end >= this->alignment.size() - 1) { break; }  // No more positions.
      this->end++;

      char c = this->alignment[this->end][this->seq];
      if(c == ' ') { continue; }
      this->context += c;
    }
  }

  Node* parent()
  {
    if(this->link != this) { this->link = this->link->parent(); }
    return this->link;
  }

  void merge(Node* another)
  {
    this->parent();
    another = another->parent();
    another->link = this->link;
  }

  void write(std::ofstream& file)
  {
    file.put(this->label);
    file.write((char*)&(this->pos), sizeof(usint));
  }
};


struct NodeContextComparator
{
  bool operator() (const Node* a, const Node* b)
  {
    if(a->label < b->label) { return true; }
    if(a->label > b->label) { return false; }

    return (a->context.compare(b->context) < 0);
  }
} node_context_comparator;


usint
writeEdges(std::vector<pair_type>& edge_vector, std::ofstream& edge_file)
{
  usint edges = 0;
  std::sort(edge_vector.begin(), edge_vector.end());
  for(usint i = 0; i < edge_vector.size(); i++)
  {
    if(i == 0 || edge_vector[i] != edge_vector[i - 1])
    {
      edge_file.write((char*)&(edge_vector[i]), sizeof(pair_type));
      edges++;
    }
  }
  return edges;
}


void
buildAutomaton(std::ifstream& input, std::ofstream& node_file, std::ofstream& edge_file, usint sequences, usint context_length, usint max_length)
{
  // Read the input file.
  input.clear();
  input.seekg(0, std::ios::beg);

  std::vector<std::string> line_buffer;
  usint lines;
  if(max_length > 0) { lines = max_length + 2; }
  else               { lines = fileSize(input) / (sequences + 1) + 2; }
  line_buffer.reserve(lines);

  line_buffer.push_back(std::string(sequences, 'Z')); // Initial line.
  while(!input.eof())
  {
    std::string buffer;
    std::getline(input, buffer);
    if(buffer.length() >= sequences)
    {
      line_buffer.push_back(buffer);
      if(max_length > 0 && line_buffer.size() >= lines) { break; }
    }
  }
  line_buffer.push_back(std::string(sequences, '\0')); // Final line.


  // Create the initial nodes.
  Node* previous_nodes[sequences];
  usint current_line = 0, nodes = 1, edges = 0;
  for(usint i = 0; i < sequences; i++)
  {
    previous_nodes[i] = new Node(line_buffer, current_line, i, context_length);
  }
  previous_nodes[0]->write(node_file);


  std::vector<pair_type> edge_vector;
  for(current_line = 1; current_line < lines; current_line++)
  {
    std::string& line = line_buffer[current_line];
    usint created = 0;

    // Build the new nodes. Do not create nodes for gaps.
    Node* current_nodes[sequences];
    for(usint i = 0; i < sequences; i++)
    {
      if(line[i] != ' ')
      {
        current_nodes[i] = new Node(previous_nodes[i], current_line);
        created++;
      }
      else { current_nodes[i] = 0; }
    }

    // Merge nodes by (label, context).
    Node* index[created]; usint previous = 0;
    for(usint i = 0; i < sequences; i++)
    {
      if(current_nodes[i] != 0) { index[previous] = current_nodes[i]; previous++; }
    }
    std::sort(index, index + created, node_context_comparator);
    previous = 0;
    for(usint i = 1; i < created; i++)
    {
      if(index[i]->label == index[previous]->label &&
         index[i]->context.compare(index[previous]->context) == 0)
      {
        index[previous]->merge(index[i]);
      }
      else { previous = i; }
    }

    // Linearize the link paths and determine the actual new nodes.
    created = 0;
    for(usint i = 0; i < sequences; i++)
    {
      if(current_nodes[i] == 0) { continue; }
      current_nodes[i]->parent();
      if(current_nodes[i]->link == current_nodes[i])
      {
        current_nodes[i]->id = nodes + created; created++;
        current_nodes[i]->write(node_file);
      }
    }
    for(usint i = 0; i < sequences; i++)
    {
      if(current_nodes[i] == 0) { continue; }
      current_nodes[i]->id = current_nodes[i]->link->id;
    }
    nodes += created;

    // Create an edge (previous_nodes[i], current_nodes[j]) if their contexts match.
    edge_vector.clear();
    for(usint j = 0; j < sequences; j++)
    {
      Node* to = current_nodes[j];
      if(to == 0) { continue; }
      std::string context; context += to->label; context += to->context;

      for(usint i = 0; i < sequences; i++)
      {
        Node* from = previous_nodes[i];
        if(current_nodes[i] == 0) { continue; } // A gap in sequence i.

        if(context.find(from->context) == 0)
        {
          edge_vector.push_back(pair_type(from->id, to->id));
        }
      }
    }

    // Output the distinct edges.
    edges += writeEdges(edge_vector, edge_file);

    // Cleanup.
    for(usint i = 0; i < sequences; i++)
    {
      if(current_nodes[i] == 0) { continue; }
      delete previous_nodes[i];
      previous_nodes[i] = current_nodes[i];
    }
  }

  std::cout << "Lines: " << lines << std::endl;
  std::cout << "Nodes: " << nodes << std::endl;
  std::cout << "Edges: " << edges << std::endl;
  std::cout << std::endl;
}


int
main(int argc, char** argv)
{
  std::cout << "Automaton builder" << std::endl;
  std::cout << std::endl;
  if(argc < 5)
  {
    std::cout << "Usage: build_automaton alignment_file node_file edge_file context_length [max_length = 0]" << std::endl;
    return 1;
  }

  std::cout << "Alignment file: " << argv[1] << std::endl;
  std::ifstream alignment_file(argv[1], std::ios_base::binary);
  if(!alignment_file)
  {
    std::cerr << "Error opening alignment file!" << std::endl;
    return 2;
  }

  std::string node_name(argv[2]);
  std::cout << "Node file: " << node_name << std::endl;
  std::ofstream node_file(node_name.c_str(), std::ios_base::binary);
  if(!node_file)
  {
    alignment_file.close();
    std::cerr << "Error creating node file!" << std::endl;
    return 3;
  }

  std::string edge_name(argv[3]);
  std::cout << "Edge file: " << edge_name << std::endl;
  std::ofstream edge_file(edge_name.c_str(), std::ios_base::binary);
  if(!edge_file)
  {
    alignment_file.close();
    node_file.close();
    std::cerr << "Error creating edge file!" << std::endl;
    return 4;
  }

  usint context_length = atoi(argv[4]);
  std::cout << "Context length: " << context_length << std::endl;

  usint max_length = 0;
  if(argc >= 6) { max_length = atoi(argv[5]); }
  std::cout << "Maximum length: " << max_length << std::endl;

  std::string line;
  std::getline(alignment_file, line);
  usint sequences = line.size();
  std::cout << "Number of sequences: " << sequences << std::endl;
  std::cout << std::endl;

  buildAutomaton(alignment_file, node_file, edge_file, sequences, context_length, max_length);
  alignment_file.close();
  node_file.close();
  edge_file.close();

#ifdef DEBUG
  Graph graph(node_name, edge_name, true);
  graph.checkGraph();
#endif

/*
  // Reorder nodes to allow more efficient sampling.
  Graph graph(node_name, edge_name, true);
  graph.reorder();
  graph.write();
*/

  return 0;
}
