#include <cstdlib>
#include <ctime>
#include <iostream>
#include <vector>

#include <misc/utils.h>

#include "gcsa.h"



using namespace CSA;


void
printUsage()
{
  std::cout << "Usage: gcsa_test [options] index [patterns]" << std::endl;
  std::cout << "  -l   Locate the occurrences." << std::endl;
  std::cout << "  -r   Search also for the reverse complements of the patterns." << std::endl;
  std::cout << std::endl;
}


int main(int argc, char** argv)
{
  std::cout << "GCSA test" << std::endl;
  std::cout << std::endl;

  bool locate = false, reverse_complement = false;
  char* index_name = 0;
  char* patterns_name = 0;
  for(int i = 1; i < argc; i++)
  {
    if(argv[i][0] == '-')
    {
      switch(argv[i][1])
      {
        case 'l':
          locate = true; break;
        case 'r':
          reverse_complement = true; break;
        default:
          std::cout << "Invalid option: " << argv[i] << std::endl << std::endl;
          printUsage();
          return 1;
      }
    }
    else if(index_name == 0)
    {
      index_name = argv[i];
    }
    else if(patterns_name == 0)
    {
      patterns_name = argv[i];
    }
  }
  if(index_name == 0) { printUsage(); return 2; }

  std::cout << "Options:";
  if(locate) { std::cout << " locate"; }
  if(reverse_complement) { std::cout << " reverse_complement"; }
  std::cout << std::endl;
  std::cout << "Index: " << index_name << std::endl;
  if(patterns_name != 0) { std::cout << "Patterns: " << patterns_name << std::endl; }
  std::cout << std::endl;

  const GCSA gcsa(index_name);
  if(!gcsa.isOk()) { return 3; }
  gcsa.reportSize(true);
  if(patterns_name == 0) { return 0; }

  std::ifstream patterns(patterns_name, std::ios_base::binary);
  if(!patterns)
  {
    std::cerr << "Error opening pattern file!" << std::endl;
    return 4;
  }

  std::vector<std::string> rows;
  readRows(patterns, rows, true);

  usint total = 0, n = rows.size();
  usint found = 0, forward = 0, reverse = 0;
  usint forward_occurrences = 0, reverse_occurrences = 0;
  double start = readTimer();
  for(usint i = 0; i < n; i++)
  {
    pair_type result = gcsa.find(rows[i]);
    total += rows[i].length();
    if(!isEmpty(result))
    {
      found++; forward++;
      if(locate)
      {
        std::vector<usint>* occurrences = gcsa.locate(result);
        if(occurrences != 0)
        {
          forward_occurrences += occurrences->size();
          delete occurrences;
        }
      }
    }

    if(reverse_complement)
    {
      pair_type reverse_result = gcsa.find(rows[i], true);
      if(!isEmpty(reverse_result))
      {
        if(isEmpty(result)) { found++; }
        reverse++;
        if(locate)
        {
          std::vector<usint>* occurrences = gcsa.locate(reverse_result);
          if(occurrences != 0)
          {
            reverse_occurrences += occurrences->size();
            delete occurrences;
          }
        }
      }
    }
  }
  double time = readTimer() - start;
  double megabases = total / (double)MILLION;

  std::cout << "Patterns:     " << n << " (" << (n / time) << " / sec)" << std::endl;

  std::cout << "Total size:   " << megabases << " megabases";
  if(!locate)
  {
    std::cout << " (" << (megabases / time) << " megabases / sec)";
  }
  std::cout << std::endl;

  std::cout << "Found:        " << found;
  if(reverse_complement)
  {
    std::cout << " (" << forward << " forward, " << reverse << " reverse complement)";
  }
  std::cout << std::endl;

  if(locate)
  {
    std::cout << "Occurrences:  " << (forward_occurrences + reverse_occurrences) << " (";
    if(reverse_complement)
    {
      std::cout << forward_occurrences << " forward, " << reverse_occurrences << " reverse complement, ";
    }
    std::cout << ((forward_occurrences + reverse_occurrences) / time) << " / sec)" << std::endl;
  }

  std::cout << "Time:         " << time << " seconds" << std::endl;
  std::cout << std::endl;

  return 0;
}
