#include <algorithm>
#include <cstdlib>
#include <sstream>
#include <queue>
#include <vector>
#include <stack>

#include "graph.h"
#include "gcsa.h"

#include <misc/utils.h>
#include <bits/deltavector.h>
#include <bits/rlevector.h>


namespace CSA
{

//--------------------------------------------------------------------------

struct GraphNodeComparator
{
  bool operator() (const GraphNode* a, const GraphNode* b)
  {
    return (a->outdegree < b->outdegree);
  }
} gn_comparator;

struct GraphEdgeComparator
{
  bool operator() (const GraphEdge& a, const GraphEdge& b)
  {
    return ((a.from < b.from) || ((a.from == b.from) && (a.to < b.to)));
  }
} ge_comparator;

struct PathNodeComparator
{
  bool operator() (const PathNode* a, const PathNode* b)
  {
    return ((a->first < b->first) || ((a->first == b->first) && (a->second < b->second)));
  }
} pn_comparator;

struct PathNodeLeftRightComparator
{
  bool operator() (const PathNode* a, const PathNode* b)
  {
    return ((a->left < b->left) || ((a->left == b->left) && (a->right < b->right)));
  }
} pn_left_right_comparator;

struct PathEdgeComparator
{
  bool operator() (const PathEdge& a, const PathEdge& b)
  {
    return ((a.from < b.from) || ((a.from == b.from) && (a.to < b.to)));
  }
} pe_comparator;

struct PathEdgeReverseComparator
{
  bool operator() (const PathEdge& a, const PathEdge& b)
  {
    return (a.to < b.to || (a.to == b.to && a.from < b.from));
  }
} pe_reverse_comparator;

struct PathEdgeFromComparator
{
  bool operator() (const PathEdge* a, const PathEdge* b)
  {
    return (a->from < b->from);
  }
} pe_from_comparator;

struct PathEdgeFullComparator
{
  bool operator() (const PathEdge* a, const PathEdge* b)
  {
    return (a->from < b->from || (a->from == b->from && a->to < b->to));
  }
} pe_full_comparator;

struct PathEdgeExtraComparator
{
  bool operator() (const PathEdge* a, const PathEdge* b)
  {
    return (a->to < b->to || (a->to == b->to && a->extra < b->extra));
  }
} pe_extra_comparator;

//--------------------------------------------------------------------------

void
deleteList(PathNodeList* list)
{
  while(list != 0)
  {
    PathNodeList* tmp = list;
    list = list->next;
    delete tmp;
  }
}

PathNode::PathNode() :
  first(0), second(0),
  sorted(false),
  from(0), to(0),
  left(0), right(0)
{
}

PathNode::~PathNode()
{
}

PathEdge::PathEdge() :
  from(0), to(0), extra(0)
{
}

PathEdge::PathEdge(PathNode* _from, PathNode* _to, PathNode* _extra) :
  from(_from), to(_to), extra(_extra)
{
}

PathNodeList::PathNodeList(PathNode* _item, PathNodeList* _next) :
  item(_item), next(_next)
{
}

bool
PathNodeList::contains(PathNode* _item)
{
  for(PathNodeList* list = this; list != 0; list = list->next)
  {
    if(list->item == _item) { return true; }
  }

  return false;
}

//--------------------------------------------------------------------------

Graph::Graph(const std::string& _node_name, const std::string& _edge_name, bool load) :
  nodes(0), edges(0),
  node_count(0), edge_count(0),
  node_name(_node_name), edge_name(_edge_name)
{
  if(!load) { return; }

  std::ifstream node_file(this->node_name.c_str(), std::ios_base::binary);
  if(!node_file)
  {
    std::cerr << "Graph: Cannot open node file (" << this->node_name << ")!" << std::endl;
    return;
  }

  std::ifstream edge_file(this->edge_name.c_str(), std::ios_base::binary);
  if(!edge_file)
  {
    std::cerr << "Graph: Cannot open edge file (" << this->edge_name << ")!" << std::endl;
    node_file.close();
    return;
  }

  this->node_count = fileSize(node_file) / (1 + sizeof(usint));
  this->nodes = new GraphNode*[this->node_count];
  for(usint i = 0; i < this->node_count; i++)
  {
    this->nodes[i] = new GraphNode;
    this->nodes[i]->label = (uchar)node_file.get();
    node_file.read((char*)&(this->nodes[i]->value), sizeof(usint));
    this->nodes[i]->paths = 0;
  }
  node_file.close();

  this->edge_count = fileSize(edge_file) / sizeof(pair_type);
  this->edges = new GraphEdge[this->edge_count];
  for(usint i = 0; i < this->edge_count; i++)
  {
    pair_type tmp;
    edge_file.read((char*)&tmp, sizeof(pair_type));
    this->edges[i].from = this->nodes[tmp.first];
    this->edges[i].to = this->nodes[tmp.second];
  }
  edge_file.close();
}

Graph::~Graph()
{
  this->deleteNodes();
  this->deleteEdges();

  this->node_count = 0; this->edge_count = 0;
}

void
Graph::deletePaths()
{
  if(this->nodes != 0)
  {
    for(usint i = 0; i < this->node_count; i++)
    {
      deleteList(this->nodes[i]->paths);
      this->nodes[i]->paths = 0;
    }
  }
}

void
Graph::deleteNodes()
{
  if(this->nodes != 0)
  {
    for(usint i = 0; i < this->node_count; i++)
    {
      deleteList(this->nodes[i]->paths);
      this->nodes[i]->paths = 0;
      delete this->nodes[i];
      this->nodes[i] = 0;
    }
  }

  delete[] this->nodes; this->nodes = 0;
}

void
Graph::deleteEdges()
{
  delete[] this->edges; this->edges = 0;
}

void
Graph::write()
{
  std::ofstream node_file(this->node_name.c_str(), std::ios_base::binary);
  if(!node_file)
  {
    std::cerr << "Graph: Cannot open node file (" << this->node_name << ")!" << std::endl;
    return;
  }

  std::ofstream edge_file(this->edge_name.c_str(), std::ios_base::binary);
  if(!edge_file)
  {
    std::cerr << "Graph: Cannot open edge file (" << this->edge_name << ")!" << std::endl;
    node_file.close();
    return;
  }

  // Nodes are written as (label, value).
  for(usint i = 0; i < this->node_count; i++)
  {
    node_file.put((uchar)(this->nodes[i]->label));
    node_file.write((char*)&(this->nodes[i]->value), sizeof(usint));
  }
  node_file.close();

  // Edges are written as (from, to).
  this->setIds();
  for(GraphEdge* ge = this->edges; ge != this->edges + this->edge_count; ge++)
  {
    edge_file.write((char*)&(ge->from->outdegree), sizeof(usint));
    edge_file.write((char*)&(ge->to->outdegree), sizeof(usint));
  }
  edge_file.close();
}

void
Graph::reorder()
{
  for(usint i = 0; i < this->node_count; i++)
  {
    this->nodes[i]->indegree = 0;
    this->nodes[i]->outdegree = 0;
  }
  for(usint i = 0; i < this->edge_count; i++) { this->edges[i].to->indegree++; }
  this->sortEdges();

  std::stack<pair_type> active;
  active.push(this->getEdgesFrom(this->nodes[0]));
  usint next_id = 1;
  while(!active.empty())
  {
    pair_type range = active.top(); active.pop();
    if(isEmpty(range)) { continue; }
    active.push(pair_type(range.first + 1, range.second));

    GraphEdge& ge = this->edges[range.first];
    ge.to->indegree--;
    if(ge.to->indegree == 0)
    {
      ge.to->outdegree = next_id; next_id++;
      active.push(this->getEdgesFrom(ge.to));
    }
  }

  std::sort(this->nodes, this->nodes + this->node_count, gn_comparator);

  #ifdef DEBUG
  for(usint i = 0; i < this->node_count; i++)
  {
    if(this->nodes[i]->outdegree != i)
    {
      std::cerr << "Graph: Node " << i << " has id " << this->nodes[i]->outdegree << "!" << std::endl;
    }
  }
  #endif
}

void
Graph::sortEdges()
{
  std::sort(this->edges, this->edges + this->edge_count, ge_comparator);
}

pair_type
Graph::getEdgesFrom(GraphNode* node)
{
  pair_type range;

  // Lower bound for the range.
  sint low = 0, high = this->edge_count - 1;
  while(low < high)
  {
    sint mid = low + (high - low) / 2;
    if(node == this->edges[mid].from)     { high = mid; }
    else if(node < this->edges[mid].from) { high = mid - 1; }
    else                                  { low = mid + 1; }
  }
  if(node == this->edges[low].from) { range.first = low; }
  else { return EMPTY_PAIR; }

  // Upper bound for the range.
  high = this->edge_count - 1;
  while(low < high)
  {
    sint mid = low + (high - low + 1) / 2;
    if(node == this->edges[mid].from) { low = mid; }
    else                              { high = mid - 1; }
  }
  range.second = high;

  return range;
}

void
Graph::checkGraph()
{
#ifdef DEBUG
  pair_type* degrees = new pair_type[this->node_count];
  usint* flags = new usint[this->node_count];
  bool* reachable = new bool[this->node_count];
  for(usint i = 0; i < this->node_count; i++)
  {
    degrees[i] = pair_type(0, 0); flags[i] = 0;
    reachable[i] = false;
  }

  // Determine predecessor labels.
  this->setIds();
  for(usint i = 0; i < this->edge_count; i++)
  {
    usint from = this->edges[i].from->outdegree;
    usint to = this->edges[i].to->outdegree;
    degrees[from].second++; degrees[to].first++;

    usint flag = 0;
    switch(this->nodes[from]->label)
    {
      case 'A':
        flag = 0x01;
        break;
      case 'C':
        flag = 0x02;
        break;
      case 'G':
        flag = 0x04;
        break;
      case 'T':
        flag = 0x08;
        break;
    }

    if(flags[to] & flag) { flag = flag | (flag << 4); }
    flags[to] |= flag;
  }

  // Determine reachable nodes.
  this->sortEdges();
  std::queue<usint> q;
  q.push(0);
  while(!q.empty())
  {
    usint i = q.front(); q.pop();
    if(reachable[i]) { continue; }
    reachable[i] = true;
    pair_type range = this->getEdgesFrom(this->nodes[i]);
    for(usint j = range.first; j <= range.second; j++) { q.push(this->edges[j].to->outdegree); }
  }

  std::cout << "Checking graph..." << std::endl;
  std::cout << std::endl;

  std::cout << "Nodes with indegree 0:" << std::endl;
  for(usint i = 0; i < this->node_count; i++)
  {
    if(degrees[i].first == 0) { std::cout << i << std::endl; }
  }
  std::cout << std::endl;

  std::cout << "Nodes with outdegree 0:" << std::endl;
  for(usint i = 0; i < this->node_count; i++)
  {
    if(degrees[i].second == 0) { std::cout << i << std::endl; }
  }
  std::cout << std::endl;

  std::cout << "Nodes with predecessors sharing a label:" << std::endl;
  for(usint i = 0; i < this->node_count; i++)
  {
    if(flags[i] & 0xF0)
    {
      std::cout << i << " (";
      if(flags[i] & 0x10) { std::cout << 'A'; }
      if(flags[i] & 0x20) { std::cout << 'C'; }
      if(flags[i] & 0x40) { std::cout << 'G'; }
      if(flags[i] & 0x80) { std::cout << 'T'; }
      std::cout << ")" << std::endl;
    }
  }
  std::cout << std::endl;

  std::cout << "Unreachable nodes:" << std::endl;
  for(usint i = 0; i < this->node_count; i++)
  {
    if(!reachable[i]) { std::cout << i << std::endl; }
  }
  std::cout << std::endl;

  std::cout << "Graph checked." << std::endl;
  std::cout << std::endl;
  delete[] degrees;
  delete[] flags;
  delete[] reachable;
#endif // DEBUG
}

void
Graph::setIds()
{
  for(usint i = 0; i < this->node_count; i++)
  {
    this->nodes[i]->outdegree = i;
  }
}

PathGraph*
Graph::getLastGeneration(bool minimal)
{
  PathGraph* pg = new PathGraph(*this, 1, false);

  while(pg->exists()) { pg->generation++; }
  if(pg->generation == 1)
  {
    delete pg;
    return 0;
  }

  pg->generation--;
  pg->load(minimal);
  pg->checkGraph();
  return pg;
}

//--------------------------------------------------------------------------

PathGraph::PathGraph(Graph& _parent, usint _generation, bool load, bool minimal) :
  nodes(0), edges(0),
  node_count(0), edge_count(0),
  parent(_parent), generation(_generation)
{
  if(load) { this->load(minimal); }
}

PathGraph::~PathGraph()
{
  this->deleteNodes();
  this->deleteEdges();

  this->node_count = 0; this->edge_count = 0;
}

void
PathGraph::deleteNodes()
{
  if(this->nodes != 0)
  {
    for(usint i = 0; i < this->node_count; i++)
    {
      delete this->nodes[i]; this->nodes[i] = 0;
    }
  }

  delete[] this->nodes; this->nodes = 0;
}

void
PathGraph::deleteEdges()
{
  delete[] this->edges; this->edges = 0;
}

std::string
PathGraph::getNodeName()
{
  std::ostringstream node_stream;
  node_stream << this->parent.node_name << "." << this->generation;
  return node_stream.str();
}

std::string
PathGraph::getEdgeName()
{
  std::ostringstream edge_stream;
  edge_stream << this->parent.edge_name << "." << this->generation;
  return edge_stream.str();
}

bool
PathGraph::exists()
{
  std::ifstream node_file(this->getNodeName().c_str(), std::ios_base::binary);
  if(!node_file) { return false; }
  node_file.close();

  std::ifstream edge_file(this->getEdgeName().c_str(), std::ios_base::binary);
  if(!edge_file) { return false; }
  edge_file.close();

  return true;
}

void
PathGraph::load(bool minimal)
{
  std::ifstream node_file(this->getNodeName().c_str(), std::ios_base::binary);
  if(!node_file)
  {
    std::cerr << "PathGraph: Cannot open node file (" << this->getNodeName() << ")!" << std::endl;
    return;
  }

  std::ifstream edge_file(this->getEdgeName().c_str(), std::ios_base::binary);
  if(!edge_file)
  {
    std::cerr << "PathGraph: Cannot open edge file (" << this->getEdgeName() << ")!" << std::endl;
    node_file.close();
    return;
  }

  this->node_count = fileSize(node_file) / (4 * sizeof(usint));
  this->nodes = new PathNode*[this->node_count];
  for(usint i = 0; i < this->node_count; i++)
  {
    this->nodes[i] = new PathNode();

    usint tmp;
    node_file.read((char*)&tmp, sizeof(usint));
    this->nodes[i]->first = tmp;
    this->nodes[i]->second = i;

    node_file.read((char*)&tmp, sizeof(usint));
    if(tmp != 0) { this->nodes[i]->sorted = true; }

    node_file.read((char*)&tmp, sizeof(usint));
    this->nodes[i]->from = this->parent.nodes[tmp];

    node_file.read((char*)&tmp, sizeof(usint));
    this->nodes[i]->to = this->parent.nodes[tmp];

    if(!minimal)
    {
      this->nodes[i]->from->paths = new PathNodeList(this->nodes[i], this->nodes[i]->from->paths);
    }
  }
  node_file.close();

  this->edge_count = fileSize(edge_file) / (2 * sizeof(usint));
  this->edges = new PathEdge[this->edge_count];
  for(usint i = 0; i < this->edge_count; i++)
  {
    usint tmp;
    edge_file.read((char*)&tmp, sizeof(usint));
    this->edges[i].from = this->nodes[tmp];
    edge_file.read((char*)&tmp, sizeof(usint));
    this->edges[i].to = this->nodes[tmp];
  }
  edge_file.close();
}

void
PathGraph::write()
{
  std::ofstream node_file(this->getNodeName().c_str(), std::ios_base::binary);
  if(!node_file)
  {
    std::cerr << "PathGraph: Cannot open node file (" << this->getNodeName() << ")!" << std::endl;
    return;
  }

  std::ofstream edge_file(this->getEdgeName().c_str(), std::ios_base::binary);
  if(!edge_file)
  {
    std::cerr << "PathGraph: Cannot open edge file (" << this->getEdgeName() << ")!" << std::endl;
    node_file.close();
    return;
  }

  // Nodes are 4-tuples (label, sorted, from, to).
  this->parent.setIds();
  for(usint i = 0; i < this->node_count; i++)
  {
    node_file.write((char*)&(this->nodes[i]->first), sizeof(usint));

    usint tmp = (this->nodes[i]->sorted ? 1 : 0);
    node_file.write((char*)&tmp, sizeof(usint));

    node_file.write((char*)&(this->nodes[i]->from->outdegree), sizeof(usint));
    node_file.write((char*)&(this->nodes[i]->to->outdegree), sizeof(usint));
  }
  node_file.close();

  // Edges are pairs (from, to). PathNode.second is assumed to contain node id.
  for(PathEdge* pe = this->edges; pe != this->edges + this->edge_count; pe++)
  {
    edge_file.write((char*)&(pe->from->second), sizeof(usint));
    edge_file.write((char*)&(pe->to->second), sizeof(usint));
  }
  edge_file.close();
}

//--------------------------------------------------------------------------

void
PathGraph::sortEdges(bool by_destination)
{
  if(by_destination)
  {
    std::sort(this->edges, this->edges + this->edge_count, pe_reverse_comparator);
  }
  else
  {
    std::sort(this->edges, this->edges + this->edge_count, pe_comparator);
  }
}

pair_type
PathGraph::getEdgesFrom(PathNode* from)
{
  pair_type range;

  // Lower bound for the range.
  sint low = 0, high = this->edge_count - 1;
  while(low < high)
  {
    sint mid = low + (high - low) / 2;
    if(from == this->edges[mid].from)     { high = mid; }
    else if(from < this->edges[mid].from) { high = mid - 1; }
    else                                  { low = mid + 1; }
  }
  if(from == this->edges[low].from) { range.first = low; }
  else { return EMPTY_PAIR; }

  // Upper bound for the range.
  high = this->edge_count - 1;
  while(low < high)
  {
    sint mid = low + (high - low + 1) / 2;
    if(from == this->edges[mid].from) { low = mid; }
    else                              { high = mid - 1; }
  }
  range.second = high;

  return range;
}

pair_type
PathGraph::getEdgesTo(PathNode* to)
{
  pair_type range;

  // Lower bound for the range.
  sint low = 0, high = this->edge_count - 1;
  while(low < high)
  {
    sint mid = low + (high - low) / 2;
    if(to == this->edges[mid].to)     { high = mid; }
    else if(to < this->edges[mid].to) { high = mid - 1; }
    else                              { low = mid + 1; }
  }
  if(to == this->edges[low].to) { range.first = low; }
  else { return EMPTY_PAIR; }

  // Upper bound for the range.
  high = this->edge_count - 1;
  while(low < high)
  {
    sint mid = low + (high - low + 1) / 2;
    if(to == this->edges[mid].to) { low = mid; }
    else                          { high = mid - 1; }
  }
  range.second = high;

  return range;
}

//--------------------------------------------------------------------------

void
PathGraph::checkGraph()
{
#ifdef DEBUG
  std::cout << "Checking path graph..." << std::endl;
  std::cout << std::endl;

  this->parent.setIds();

  pair_type* degrees = new pair_type[this->node_count + 1];
  bool* self = new bool[this->node_count];
  bool* reachable = new bool[this->node_count];
  for(usint i = 0; i < this->node_count; i++)
  {
    degrees[i] = pair_type(0, 0); self[i] = false;
    reachable[i] = false;
  }

  for(usint i = 0; i < this->edge_count; i++)
  {
    degrees[this->edges[i].from->second].second++;
    degrees[this->edges[i].to->second].first++;
    if(this->edges[i].from == this->edges[i].to) { self[i] = true; }
  }

  for(usint l = this->edge_count; l > 0; l--)
  {
    usint i = l - 1;
    if(reachable[this->edges[i].from->second]) { reachable[this->edges[i].to->second] = true; }
  }

  std::cout << "Nodes with indegree 0:" << std::endl;
  for(usint i = 0; i < this->node_count; i++)
  {
    if(degrees[i].first == 0) { this->printNodeInfo(i); }
  }
  std::cout << std::endl;

  std::cout << "Nodes with outdegree 0:" << std::endl;
  for(usint i = 0; i < this->node_count; i++)
  {
    if(degrees[i].second == 0) { this->printNodeInfo(i); }
  }
  std::cout << std::endl;

  for(usint i = 0; i < this->node_count; i++) { degrees[i] = EMPTY_PAIR; }
  this->sortEdges();
  for(usint i = 0; i < this->edge_count; i++)
  {
    usint from = this->edges[i].from->second;
    if(degrees[from] == EMPTY_PAIR) { degrees[from] = pair_type(i, i); }
  }
  degrees[this->node_count] = pair_type(this->edge_count, this->edge_count);
  std::queue<usint> q;
  q.push(this->node_count - 1);
  while(!q.empty())
  {
    usint i = q.front(); q.pop();
    if(reachable[i]) { continue; }
    reachable[i] = true;
    for(usint j = degrees[i].first; j < degrees[i + 1].first; j++)
    {
      q.push(this->edges[j].to->second);
    }
  }

  std::cout << "Nodes with self-cycles:" << std::endl;
  for(usint i = 0; i < this->node_count; i++)
  {
    if(self[i]) { this->printNodeInfo(i); }
  }
  std::cout << std::endl;

  std::cout << "Unreachable nodes:" << std::endl;
  for(usint i = 0; i < this->node_count; i++)
  {
    if(!reachable[i]) { this->printNodeInfo(i); }
  }
  std::cout << std::endl;

  std::cout << "Path graph checked." << std::endl;
  std::cout << std::endl;

  delete[] degrees;
  delete[] self;
  delete[] reachable;
#endif // DEBUG
}

void
PathGraph::printNodeInfo(usint i)
{
  usint from = this->nodes[i]->from->outdegree;
  usint to = this->nodes[i]->to->outdegree;

  std::cout << "  " << i << " (label " << this->nodes[i]->first << ", from " << from << " to " << to << ")" << std::endl;
}

//--------------------------------------------------------------------------

std::vector<pair_type>*
PathGraph::getSamples(usint sample_rate)
{
  std::vector<pair_type>* pairs = new std::vector<pair_type>();
  pairs->reserve((this->node_count / sample_rate) + (this->edge_count - this->node_count + 1));

  bool* found = new bool[this->node_count];
  for(usint i = 0; i < this->node_count; i++)
  {
    found[i] = false;
    this->nodes[i]->second = i;
  }

  this->sortEdges();
  std::stack<PathNode*> active;
  active.push(this->nodes[this->node_count - 1]); // Initial node.
  usint unsampled = 0;
  PathNode* current = 0;
  while(!active.empty())
  {
    current = active.top(); active.pop();
    unsampled = 1;

    while(true)
    {
      if(found[current->second]) { break; }
      found[current->second] = true;
      pair_type range = this->getEdgesFrom(current);

      // Sample if:
      //   i)   there are no samples within sample_rate steps;
      //   ii)  current node does not have exactly one outgoing edge; or
      //   iii) (value of successor) is not (value of current node + 1).
      if(unsampled >= sample_rate ||
         length(range) != 1 ||
         this->edges[range.first].to->from->value != current->from->value + 1)
      {
        pairs->push_back(pair_type(current->second, current->from->value));
        for(usint i = range.first; i <= range.second; i++)
        {
          active.push(this->edges[i].to);
        }
        break;
      }

      current = this->edges[range.first].to;
      unsampled++;
    }
  }

  std::sort(pairs->begin(), pairs->end());

  return pairs;
}

//--------------------------------------------------------------------------

bool
updateLabels(PathNode** pathnodes, usint nodes)
{
  usint rank = 0;
  bool sorted = true;
  pair_type prev(nodes, nodes);   // Previous (first, second) values.
  PathNode* start = 0;            // First PathNode with given (first, second) values.

  std::sort(pathnodes, pathnodes + nodes, pn_comparator);

  for(usint i = 0; i < nodes; i++)
  {
    if(pathnodes[i]->first != prev.first || pathnodes[i]->second != prev.second)
    {
      rank++;
      prev.first = pathnodes[i]->first; prev.second = pathnodes[i]->second;
      if(start != 0)
      {
        start->sorted = true;
      }
      start = pathnodes[i];
    }
    else
    {
      sorted = false;
      start = 0;
    }
    pathnodes[i]->first = rank;
    pathnodes[i]->second = i; // A permanent id for the node.
  }
  if(start != 0)
  {
    start->sorted = true;
  }

  std::cout << "Path labels: " << rank << std::endl;

  return sorted;
}

PathNode*
createPathNode(GraphNode* from, GraphNode* to)
{
  PathNode* pn = new PathNode();

  pn->first = from->label;
  if(to != from)
  {
    pn->second = to->label;
  }
  pn->from = from; pn->to = to;
  pn->from->paths = new PathNodeList(pn, pn->from->paths);

  return pn;
}

//--------------------------------------------------------------------------

/*
  Note that we can create multiple PathNodes with from == initial node.
  They will be merged in the first pruning step, because they are the only nodes
  with path label starting with 'Z'.
*/
bool
initPaths(Graph& graph)
{
  usint pn_count = graph.edge_count + 1, pe_count = 0;
  PathNode** pathnodes = new PathNode*[pn_count];

  // Each edge becomes a path of length 2.
  for(usint i = 0; i < graph.edge_count; i++)
  {
    pathnodes[i] = createPathNode(graph.edges[i].from, graph.edges[i].to);
  }

  // A new path containing just the end marker. Already sorted.
  pathnodes[graph.edge_count] = createPathNode(graph.nodes[graph.node_count - 1],
                                               graph.nodes[graph.node_count - 1]);
  pathnodes[graph.edge_count]->sorted = true;
  pn_count = graph.edge_count + 1;

  // Determine the number of PathEdges.
  for(usint i = 0; i < pn_count - 1; i++)
  {
    for(PathNodeList* suc = pathnodes[i]->to->paths; suc != 0; suc = suc->next)
    {
      pe_count++;
    }
  }

  // Determine the PathEdges.
  PathEdge* pathedges = new PathEdge[pe_count];
  for(usint i = 0, j = 0; i < pn_count - 1; i++)
  {
    for(PathNodeList* suc = pathnodes[i]->to->paths; suc != 0; suc = suc->next)
    {
      pathedges[j].from = pathnodes[i];
      pathedges[j].to = suc->item;
      j++;
    }
  }

  // Sort PathNodes, update labels, and write PathGraph.
  bool sorted = updateLabels(pathnodes, pn_count);
  PathGraph pg(graph, 1, false);
  pg.nodes = pathnodes; pg.node_count = pn_count;
  pg.edges = pathedges; pg.edge_count = pe_count;
  pg.write();

  std::cout << "Nodes: " << graph.node_count << " -> " << pg.node_count << std::endl;
  std::cout << "Edges: " << graph.edge_count << " -> " << pg.edge_count << std::endl;
  std::cout << std::endl;

  return sorted;
}

//--------------------------------------------------------------------------

PathNode*
mergePathNodes(PathNode* left, PathNode* right)
{
  PathNode* pn = new PathNode();

  pn->first = left->first;
  pn->from = left->from;
  pn->left = left;

  if(right != 0)
  {
    pn->second = right->first;
    pn->sorted = left->sorted | right->sorted;
    pn->to = right->to;
    pn->right = right;
  }
  else
  {
    pn->sorted = left->sorted;
    pn->to = left->to;
  }

  return pn;
}

pair_type
getNodesLeft(PathNode* left, PathNode** nodes, sint node_count)
{
  pair_type range;

  // Lower bound for the range.
  sint low = 0, high = node_count - 1;
  while(low < high)
  {
    sint mid = low + (high - low) / 2;
    if(left == nodes[mid]->left)     { high = mid; }
    else if(left < nodes[mid]->left) { high = mid - 1; }
    else                             { low = mid + 1; }
  }
  if(left == nodes[low]->left) { range.first = low; }
  else { return EMPTY_PAIR; }

  // Upper bound for the range.
  high = node_count - 1;
  while(low < high)
  {
    sint mid = low + (high - low + 1) / 2;
    if(left == nodes[mid]->left) { low = mid; }
    else                         { high = mid - 1; }
  }
  range.second = high;

  return range;
}

pair_type
getEdgesFrom(PathNode* from, PathEdge** edges, sint edge_count)
{
  pair_type range;

  // Lower bound for the range.
  sint low = 0, high = edge_count - 1;
  while(low < high)
  {
    sint mid = low + (high - low) / 2;
    if(from == edges[mid]->from)     { high = mid; }
    else if(from < edges[mid]->from) { high = mid - 1; }
    else                             { low = mid + 1; }
  }
  if(from == edges[low]->from) { range.first = low; }
  else { return EMPTY_PAIR; }

  // Upper bound for the range.
  high = edge_count - 1;
  while(low < high)
  {
    sint mid = low + (high - low + 1) / 2;
    if(from == edges[mid]->from) { low = mid; }
    else                         { high = mid - 1; }
  }
  range.second = high;

  return range;
}


// Finds PathEdges (from, to->left, to->right).
pair_type
getEdgesTo(PathNode* to, PathEdge** edges, sint edge_count)
{
  pair_type range;

  // Lower bound for the range.
  sint low = 0, high = edge_count - 1;
  while(low < high)
  {
    sint mid = low + (high - low) / 2;
    if(to->left == edges[mid]->to && to->right == edges[mid]->extra)
    {
      high = mid;
    }
    else if(to->left < edges[mid]->to || (to->left == edges[mid]->to && to->right < edges[mid]->extra))
    {
      high = mid - 1;
    }
    else { low = mid + 1; }
  }
  if(to->left == edges[low]->to && to->right == edges[low]->extra) { range.first = low; }
  else { return EMPTY_PAIR; }

  // Upper bound for the range.
  high = edge_count - 1;
  while(low < high)
  {
    sint mid = low + (high - low + 1) / 2;
    if(to->left == edges[mid]->to && to->right == edges[mid]->extra) { low = mid; }
    else { high = mid - 1; }
  }
  range.second = high;

  return range;
}

// Returns the next maximal mergeable set of PathNodes.
// A set of PathNodes sharing adjacent path labels (first values) is mergeable,
// if each of the PathNodes begins in the same GraphNode.
// If the maximal set is empty, returns the next PathNode.
pair_type
nextMaximalSet(pair_type range, PathNode** pathnodes, usint node_count)
{
  if(range.second >= node_count - 1) { return EMPTY_PAIR; }

  if(isEmpty(range))
  {
    range.first = 0;
    range.second = 0;
  }
  else
  {
    range.first = range.second + 1;
    range.second = range.first;
    if(pathnodes[range.first - 1]->first == pathnodes[range.first]->first) { return range; }
  }

  for(usint i = range.second + 1; i < node_count; i++)
  {
    if(pathnodes[i]->first != pathnodes[i - 1]->first) { range.second = i - 1; }
    if(pathnodes[i]->from != pathnodes[range.first]->from) { return range; }
  }

  range.second = node_count - 1;
  return range;
}

//--------------------------------------------------------------------------

bool
doublePaths(Graph& graph)
{
  PathGraph* pg = graph.getLastGeneration();
  if(pg == 0)
  {
    std::cerr << "Error: No PathGraph exists!" << std::endl;
    return false;
  }
  std::cout << "Generation: " << pg->generation << " -> " << (pg->generation + 1) << std::endl;
  std::cout << std::endl;


  // Part 1: Determine the new PathNodes. OK

  PathEdge** pathedges = new PathEdge*[pg->edge_count];
  for(usint i = 0; i < pg->edge_count; i++) { pathedges[i] = &(pg->edges[i]); }
  std::sort(pathedges, pathedges + pg->edge_count, pe_from_comparator);

  std::vector<PathNode*>* pn_vector = new std::vector<PathNode*>();
  pn_vector->reserve(pg->node_count + pg->node_count / 4); // Overestimate?
  std::vector<PathEdge*>* pe_vector = new std::vector<PathEdge*>();
  pe_vector->reserve(pg->edge_count + pg->edge_count / 4); // Overestimate?

  graph.sortEdges();

  usint case1 = 0, case2 = 0, case3 = 0;
  for(usint i = 0; i < pg->node_count; i++)
  {
    PathNode* left = pg->nodes[i]; // left denotes path from A to C.

    pair_type pe_range = getEdgesFrom(left, pathedges, pg->edge_count);

    // Case 1: If AC is already sorted, it survives into next generation.
    // There will be a PathEdge from AC to every PathNode beginning with BD.
    if(left->sorted)
    {
      PathNode* temp = mergePathNodes(left, 0);
      pn_vector->push_back(temp);
      for(usint j = pe_range.first; j <= pe_range.second; j++)
      {
        PathNode* lsuc = pathedges[j]->to;  // lsuc denotes path from B to D.
        case1++;
        pe_vector->push_back(new PathEdge(temp, lsuc, 0));
      }
      continue;
    }

    // Iterate through the nodes D that are successors of C.
    pair_type ge_range = graph.getEdgesFrom(left->to);
    for(usint ge = ge_range.first; ge <= ge_range.second; ge++)
    {
      GraphNode* nodeD = graph.edges[ge].to;

      // We create a new PathNode ACDF for every old PathNode DF.
      for(PathNodeList* list = nodeD->paths; list != 0; list = list->next)
      {
        PathNode* right = list->item; // right denotes path from D to F.
        PathNode* temp = mergePathNodes(left, right);
        pn_vector->push_back(temp);

        // Iterate through all successors BD of AC.
        for(usint j = pe_range.first; j <= pe_range.second; j++)
        {
          PathNode* lsuc = pathedges[j]->to;  // lsuc denotes path from B to D.

          // Ignore this successor if it does not end at D.
          // Note that if lsuc is already sorted, it might contain a path with to == D.
          // But in this case, ACDF will become sorted, as ACD is already unique.
          // Also ACD'F' will be sorted, where D' == lsuc->to, and merged with ACDF.
          // If lsuc does not contain any proper extensions of AC, then it is sorted,
          // and hence AC is.
          if(lsuc->to != nodeD) { continue; }

          // Case 2: If BD is already sorted, then there will be a PathEdge from ACDF to BD.
          if(lsuc->sorted)
          {
            pe_vector->push_back(new PathEdge(temp, lsuc, 0));
            case2++;
            continue;
          }

          // Case 3: There will be a PathEdge from ACDF to BDEG for every old PathEdge from DF to EG.
          pair_type rng = getEdgesFrom(right, pathedges, pg->edge_count);
          for(usint k = rng.first; k <= rng.second; k++)
          {
            PathNode* rsuc = pathedges[k]->to;  // rsuc denotes path from E to G.
            case3++;
            pe_vector->push_back(new PathEdge(temp, lsuc, rsuc));
          }
        }
      }
    }
  }

  // Free some memory.
  graph.deletePaths();
  graph.deleteEdges();

  usint pn_count = pn_vector->size();
  std::cout << "Created " << pn_count << " nodes." << std::endl;
#ifdef DEBUG
  std::cout << "Initialized edges:" << std::endl;
  std::cout << "  Case 1: " << case1 << std::endl;
  std::cout << "  Case 2: " << case2 << std::endl;
  std::cout << "  Case 3: " << case3 << std::endl;
#endif
  std::cout << std::endl;


  // Part 2: Determine the new PathEdges. OK

  PathNode** pathnodes = new PathNode*[pn_count];
  for(usint i = 0; i < pn_count; i++) { pathnodes[i] = pn_vector->at(i); }
  std::sort(pathnodes, pathnodes + pn_count, pn_left_right_comparator);
  delete pn_vector; pn_vector = 0;

  usint pe_count = pe_vector->size();
  delete[] pathedges; pathedges = new PathEdge*[pe_count];
  for(usint i = 0; i < pe_count; i++) { pathedges[i] = pe_vector->at(i); }
  std::sort(pathedges, pathedges + pe_count, pe_extra_comparator);
  delete pe_vector; pe_vector = new std::vector<PathEdge*>();
  pe_vector->reserve(pe_count + pe_count / 4); // Overestimate?

  usint done1 = 0, done2 = 0, done3 = 0;

  // Cases 1 and 2: Replace (from, to->left, 0) by from, to for every to.
  for(usint i = 0; i < pe_count; i++)
  {
    PathEdge* pe = pathedges[i];
    if(pe->extra != 0) { continue; }

    PathNode* left = pe->to;
    pair_type range = getNodesLeft(left, pathnodes, pn_count);
    for(usint j = range.first; j <= range.second; j++)
    {
      if(pe->from->left->sorted) { done1++; } else { done2++; }
      pe_vector->push_back(new PathEdge(pe->from, pathnodes[j]));
    }
  }

  // Case 3: Replace (from, to->left, to->right) by (from, to) unless to->right == 0.
  for(usint i = 0; i < pn_count; i++)
  {
    PathNode* to = pathnodes[i];
    if(to->right == 0) { continue; }

    // PathEdges (from, to) are stored as (from, to->left, to->right).
    pair_type range = getEdgesTo(to, pathedges, pe_count);
    for(usint j = range.first; j <= range.second; j++)
    {
      done3++;
      pe_vector->push_back(new PathEdge(pathedges[j]->from, to));
      pathedges[j]->from = 0;
    }
  }

  std::cout << "Created " << pe_vector->size() << " edges." << std::endl;
#ifdef DEBUG
  usint unused = 0;
  for(usint i = 0; i < pe_count; i++)
  {
    PathEdge* pe = pathedges[i];
    if(pe->from != 0 && pe->extra != 0) { unused++; }
  }

  std::cout << "Edges (created / initialized):" << std::endl;
  std::cout << "  Case 1: " << done1 << " / " << case1 << std::endl;
  std::cout << "  Case 2: " << done2 << " / " << case2 << std::endl;
  std::cout << "  Case 3: " << done3 << " / " << case3 << std::endl;
  std::cout << "  Unused: " << unused << std::endl;
#endif
  std::cout << std::endl;


  // Part 3: Get rid of the old path graph. OK
  pg->deleteNodes();
  pg->deleteEdges();


  // Part 4: Sort the new PathNodes and update their labels. OK.
  std::cout << "Doubling completed." << std::endl;
  bool sorted = updateLabels(pathnodes, pn_count);
  std::cout << "Nodes: " << pg->node_count << " -> " << pn_count << std::endl;
  std::cout << "Edges: " << pg->edge_count << " -> " << pe_vector->size() << std::endl;
  std::cout << std::endl;


  // Part 5: Merge PathNodes when possible. OK

  for(usint i = 0; i < pe_count; i++) { delete pathedges[i]; }
  pe_count = pe_vector->size();
  delete[] pathedges; pathedges = new PathEdge*[pe_count];
  for(usint i = 0; i < pe_count; i++) { pathedges[i] = pe_vector->at(i); }
  delete pe_vector; pe_vector = 0;

  // Find the maximal mergeable sets.
  pair_type node_range = EMPTY_PAIR;
  while(true)
  {
    node_range = nextMaximalSet(node_range, pathnodes, pn_count);
    if(isEmpty(node_range)) { break; }

    PathNode* representative = pathnodes[node_range.first];
    for(usint j = node_range.first; j <= node_range.second; j++)
    {
      pathnodes[j]->left = representative;
    }
  }

  // Update PathEdges and remove duplicates.
  for(usint i = 0; i < pe_count; i++)
  {
    pathedges[i]->from = pathedges[i]->from->left;
    pathedges[i]->to = pathedges[i]->to->left;
  }
  std::sort(pathedges, pathedges + pe_count, pe_full_comparator);
  usint j = 0;
  for(usint i = 1; i < pe_count; i++)
  {
    if(pathedges[i]->from != pathedges[j]->from || pathedges[i]->to != pathedges[j]->to)
    {
      j++; pathedges[j] = pathedges[i];
    }
    else { delete pathedges[i]; }
  }
  pe_count = j + 1;

  // Remove nodes that are not marked as representatives of the mergeable set.
  j = 0;
  for(usint i = 1; i < pn_count; i++)
  {
    if(pathnodes[i]->left == pathnodes[i])
    {
      j++; pathnodes[j] = pathnodes[i];
    }
    else
    {
      delete pathnodes[i];
    }
  }
  pn_count = j + 1;


  // Part 6: Update labels. OK
  std::cout << "Pruning completed." << std::endl;
  for(usint i = 0; i < pn_count; i++) { pathnodes[i]->second = 0; }
  sorted = updateLabels(pathnodes, pn_count);
  std::cout << "Nodes: " << pg->node_count << " -> " << pn_count << std::endl;
  std::cout << "Edges: " << pg->edge_count << " -> " << pe_count << std::endl;
  std::cout << std::endl;


  // Finally: Build and write the generation n+1 PathGraph. OK

  usint generation = pg->generation + 1;
  delete pg; pg = 0;

  PathEdge* new_pathedges = new PathEdge[pe_count];
  for(usint i = 0; i < pe_count; i++)
  {
    new_pathedges[i].from = pathedges[i]->from;
    new_pathedges[i].to = pathedges[i]->to;
    delete pathedges[i];
  }
  delete pathedges; pathedges = 0;

  PathGraph doubled(graph, generation, false);
  doubled.nodes = pathnodes; doubled.node_count = pn_count;
  doubled.edges = new_pathedges; doubled.edge_count = pe_count;
  doubled.write();

  return sorted;
}

//--------------------------------------------------------------------------

} // namespace CSA
