#include <cstdlib>
#include <ctime>
#include <iostream>
#include <vector>

#include <rlcsa.h>

#include "gcsa.h"



using namespace CSA;


void
printUsage()
{
  std::cout << "Usage: rlcsa_test [options] index [patterns]" << std::endl;
  std::cout << "  -l   Locate the occurrences." << std::endl;
  std::cout << "  -r   Search also for the reverse complements of the patterns." << std::endl;
  std::cout << "  -w   Write found patterns into different files." << std::endl;
  std::cout << std::endl;
}

char
complement(char c)
{
  usint temp = GCSA::ALPHABET.find(c);
  if(temp != std::string::npos) { c = GCSA::COMPLEMENT[temp]; }
  return c;
}

std::string&
reverseComplement(std::string& s)
{
  usint n = s.length();
  for(usint i = 0, j = n - 1; i < j; i++, j--)
  {
    char temp = complement(s[i]);
    s[i] = complement(s[j]);
    s[j] = temp;
  }

  return s;
}

class PatternClassifier
{
  public:
    PatternClassifier(const std::string& base_name) :
      classify(false),
      forward_file(0), reverse_file(0), notfound_file(0)
    {
      if(base_name.length() == 0) { return; }

      std::string forward_name = base_name + ".forward";
      this->forward_file = new std::ofstream(forward_name.c_str(), std::ios_base::binary);
      if(!*(this->forward_file)) { return; }

      std::string reverse_name = base_name + ".reverse";
      this->reverse_file = new std::ofstream(reverse_name.c_str(), std::ios_base::binary);
      if(!*(this->reverse_file)) { return; }

      std::string notfound_name = base_name + ".notfound";
      this->notfound_file = new std::ofstream(notfound_name.c_str(), std::ios_base::binary);
      if(!*(this->notfound_file)) { return; }

      this->classify = true;
    }

    ~PatternClassifier()
    {
      delete this->forward_file; this->forward_file = 0;
      delete this->reverse_file; this->reverse_file = 0;
      delete this->notfound_file; this->notfound_file = 0;
    }

    void forward(std::string& pattern)
    {
      if(this->classify)
      {
        *(this->forward_file) << pattern << std::endl;
      }
    }

    void reverse(std::string& pattern)
    {
      if(this->classify)
      {
        *(this->reverse_file) << pattern << std::endl;
      }
    }

    void notfound(std::string& pattern)
    {
      if(this->classify)
      {
        *(this->notfound_file) << pattern << std::endl;
      }
    }

  private:
    bool classify;
    std::ofstream* forward_file;
    std::ofstream* reverse_file;
    std::ofstream* notfound_file;
};



int main(int argc, char** argv)
{
  std::cout << "RLCSA test" << std::endl;
  std::cout << std::endl;

  bool locate = false, reverse_complement = false, write = false;
  char* index_name = 0;
  char* patterns_name = 0;
  for(int i = 1; i < argc; i++)
  {
    if(argv[i][0] == '-')
    {
      switch(argv[i][1])
      {
        case 'l':
          locate = true; break;
        case 'r':
          reverse_complement = true; break;
        case 'w':
          write = true; break;
        default:
          std::cout << "Invalid option: " << argv[i] << std::endl << std::endl;
          printUsage();
          return 1;
      }
    }
    else if(index_name == 0)
    {
      index_name = argv[i];
    }
    else if(patterns_name == 0)
    {
      patterns_name = argv[i];
    }
  }
  if(index_name == 0) { printUsage(); return 2; }

  std::cout << "Options:";
  if(locate) { std::cout << " locate"; }
  if(reverse_complement) { std::cout << " reverse_complement"; }
  if(write) { std::cout << " write"; }
  std::cout << std::endl;
  std::cout << "Index: " << index_name << std::endl;
  if(patterns_name != 0) { std::cout << "Patterns: " << patterns_name << std::endl; }
  std::cout << std::endl;

  const RLCSA rlcsa(index_name);
  if(!rlcsa.isOk()) { return 3; }
  rlcsa.reportSize(true);
  if(patterns_name == 0) { return 0; }

  std::ifstream patterns(patterns_name, std::ios_base::binary);
  if(!patterns)
  {
    std::cerr << "Error opening pattern file!" << std::endl;
    return 4;
  }
  PatternClassifier classifier(write ? patterns_name : "");

  std::vector<std::string> rows;
  readRows(patterns, rows, true);
  usint total = 0, n = rows.size();
  usint found = 0, forward = 0, reverse = 0;
  usint forward_occurrences = 0, reverse_occurrences = 0;

  double start = readTimer();
  for(usint i = 0; i < n; i++)
  {
    pair_type result = rlcsa.count(rows[i]);
    total += rows[i].length();
    if(!isEmpty(result))
    {
      found++; forward++;
      if(write) { classifier.forward(rows[i]); }
      if(locate)
      {
        usint* occurrences = rlcsa.locate(result);
        for(usint j = 0; j < length(result); j++)
        {
          pair_type tmp = rlcsa.getRelativePosition(occurrences[j]);
        }
        forward_occurrences += length(result);
        delete[] occurrences;
      }
    }

    if(reverse_complement)
    {
      reverseComplement(rows[i]);
      pair_type reverse_result = rlcsa.count(rows[i]);
      if(!isEmpty(reverse_result))
      {
        if(isEmpty(result)) { found++; }
        reverse++;
        if(write){ classifier.reverse(reverseComplement(rows[i])); }
        if(locate)
        {
          usint* occurrences = rlcsa.locate(reverse_result);
          for(usint j = 0; j < length(reverse_result); j++)
          {
            rlcsa.getRelativePosition(occurrences[j]);
          }
          reverse_occurrences += length(reverse_result);
          delete[] occurrences;
        }
      }
      else if(isEmpty(result) && write)
      {
        classifier.notfound(reverseComplement(rows[i]));
      }
    }
  }
  double time = readTimer() - start;
  double megabases = total / (double)MILLION;

  std::cout << "Patterns:     " << n << " (" << (n / time) << " / sec)" << std::endl;

  std::cout << "Total size:   " << megabases << " megabases";
  if(!locate)
  {
    std::cout << " (" << (megabases / time) << " megabases / sec)";
  }
  std::cout << std::endl;

  std::cout << "Found:        " << found;
  if(reverse_complement)
  {
    std::cout << " (" << forward << " forward, " << reverse << " reverse complement)";
  }
  std::cout << std::endl;

  if(locate)
  {
    std::cout << "Occurrences:  " << (forward_occurrences + reverse_occurrences) << " (";
    if(reverse_complement)
    {
      std::cout << forward_occurrences << " forward, " << reverse_occurrences << " reverse complement, ";
    }
    std::cout << ((forward_occurrences + reverse_occurrences) / time) << " / sec)" << std::endl;
  }

  std::cout << "Time:         " << time << " seconds" << std::endl;
  std::cout << std::endl;

  return 0;
}
